--basic.lua
--These are all basic objects and functions the game needs
--Most of these will be setup at startup or firsttime setup
--BasicData object contains all 'basic' functions
--Made to be reusable

local BasicData = {}

--BASIC

--Sprite: A spritesheet is spliced by y,x into frames. These frames are put into a 2D array.
-- You call a quad with its x and y coord, and then fill in the o
--Created through databse
Sprite = Object:extend()
function Sprite:new(imgPath,spriteWidth,spriteHeight)
	if spriteWidth <= 0 then spriteWidth = 1 end
	if spriteHeight <= 0 then spriteHeight = 1 end
	--BasicData Setup
	self.width = spriteWidth
	self.height = spriteHeight
	self.img = nil
	self.quads = {}
	self.columns = 1
	self.rows = 1
	if love.filesystem.getInfo(imgPath) then
		self.img = love.graphics.newImage(imgPath)
		--Creating quads
		self.columns = math.floor(self.img:getWidth()/self.width)
		self.rows = math.floor(self.img:getHeight()/self.height)
		for y=0,self.rows-1 do
			for x=0,self.columns-1 do
				local quad = love.graphics.newQuad(x*spriteWidth,y*spriteHeight,spriteWidth,spriteHeight,self.img)
				table.insert(self.quads,quad)
			end
		end
	end
end
function Sprite:getFrame(id)
	return self.quads[id]
end
function Sprite:draw(index,x,y,a,sx,sy,ox,oy)
	
	--Defaults
	if a == nil then a = 0 end
	if sx == nil then sx = 1 end
	if sy == nil then sy = 1 end
	if ox == nil then ox = 0 end
	if oy == nil then oy = 0 end
	--The actual drawing
	if self.img == nil then 
		love.graphics.print("NaN",math.ceil(x),math.ceil(y))
	else 
		if self.quads[index] ~= nil then
			love.graphics.draw(self.img,self:getFrame(index),math.ceil(x),math.ceil(y),a,sx,sy,ox,oy)
		else
			love.graphics.draw(self.img,math.ceil(x),math.ceil(y),a,sx,sy,ox,oy)
		end
	end
end


--Animation: Contains a string, as sprite, and an array of frame ids (ints)
--also has an array pointer. Updates per drawing, unless paused
--Has an animation speed
--If the frame has a -1, do not draw!
--Created through database
Animation = Object:extend()
function Animation:new(myName, mySprite, myFrames)
	self.name = myName
	self.sprite = mySprite
	self.frames = myFrames
	self.index = 1
	self.spd = 1.0
	self.is_paused = false
	self.loops = false
	self.angle = 0
	self.scale = {x=1,y=1}
	self.origin = {x=0,y=0}
	self.visible = true
end
function Animation:setVisible(b)
	if b == nil then b = false end
	self.visible = b
end
function Animation:setIndex(i)
	self.index = Lume.clamp(i,1,#self.frames)
end
function Animation:randomIndex()
	if #self.frames > 0 then
		self.index = love.math.random(1,#self.frames)
	end
end
function Animation:setOrigin(x,y)
	self.origin.x = x
	self.origin.y = y
end
function Animation:setAngle(r)
	self.angle = r
	if self.angle < 0 then self.angle = 0 
	elseif self.angle > 2 * math.pi then self.angle = 2 * math.pi
	end
end
function Animation:setScale(x,y)
	self.scale.x = x
	self.scale.y = y
end
function Animation:update(dt)
	if self.is_paused == false then
		self.index = self.index + (dt * self.spd)
		if self.index >= #self.frames+1 then self.index = 1 end
	end	
end
function Animation:draw(x,y)
	if self.visible then
		local frame = self.frames[math.floor(self.index)]
		if frame > 0 then 
			self.sprite:draw(frame,x,y,self.angle,self.scale.x,self.scale.y,self.origin.x,self.origin.y) 
		end
	end
end
function Animation:setSpeed(i)
	if i == nil then i = 1 end
	self.spd = i
end


--TileSet: Tiles used for mapping
--Each have a name and a bunch of Tiles
--The Tilemap will use a spritebatch with the quads
--But it does draw them if needed
--Created through database.
TileSet = Object:extend()
BasicData.Tilesets = {}
function TileSet:new(name,imgPath,width,height)
	self.name = name
	self.tileSize = {w = width,h = height}
	self.img = nil
	self.tiles = {}
	self.columns = 1
	self.rows = 1
	if love.filesystem.getInfo(imgPath) then
		self.img = love.graphics.newImage(imgPath)
		self.columns = math.floor(self.img:getWidth()/width)
		self.rows = math.floor(self.img:getHeight()/height)
		for y=0,self.rows-1 do
			for x=0,self.columns-1 do
				local quad = love.graphics.newQuad(x*width,y*height,width,height,self.img)
				table.insert(self.tiles,quad)
			end
		end
	end
	table.insert(BasicData.Tilesets,self)
	self.id = #BasicData.Tilesets
end
function TileSet:getTile(id)
	id = Lume.clamp(id,1,#self.tiles)
	return self.tiles[id]
end
-- Just draw one tile
function TileSet:draw(id,x,y)
	love.graphics.draw(self.img,self.tile[id],x,y)
end
function BasicData.getTileSet(id)
	if id == nil then return nil end
	id = Lume.clamp(id,1,#BasicData.Tilesets)
	return BasicData.Tilesets[id]
end

--Camera: controls what is shown on the canvas
Camera = Object:extend()
function Camera:new(offsetx,offsety)
	self.offset = {x=offsetx,y=offsety}
	self.scale = {x=1,y=1}
	self.pixelPerfect = false 
	self.shaketime = 0.0
	self.shakepower = 1
end
function Camera:setOffset(x,y)
	self.offset.x = Lume.clamp(x,0,1)
	self.offset.y = Lume.clamp(y,0,1)
end
function Camera:setScale(x,y)
	self.scale.x = x
	self.scale.y = y
end
function Camera:setScreenShake(duration,power)
	--Defaults
	if duration == nil then duration = 1 end
	if power == nil then power = 1 end
	--Setting
	self.shaketime = math.abs(duration)
	self.shakepower = math.abs(power)
	
	self.shakeOffset = {x=0,y=0}
end
function Camera:process(dt)
	if self.shaketime > 0 then
		self.shaketime = self.shaketime - dt
		self.shakeOffset.x = math.random(-self.shakepower,self.shakepower)*dt
		self.shakeOffset.y = math.random(-self.shakepower,self.shakepower)*dt
	end
end
function Camera:draw(x,y)
	local a,b = love.graphics.getDimensions()
	local xpos = -x + (self.offset.x*a)
	local ypos = -y + (self.offset.y*b)
	if self.pixelPerfect then
		love.graphics.translate(math.floor(xpos/self.scale.x),math.floor(ypos/self.scale.y))
	else
		love.graphics.translate(xpos,ypos)
	end
	--Screenshake
	if self.shaketime > 0 then
		love.graphics.translate(self.shakeOffset.x,self.shakeOffset.y)
	end
	love.graphics.scale(self.scale.x,self.scale.y)
end 

--Shape: Used for collisions and rays.
--They also have an optional draw function in case you need to see a shape drawn
--Has a bunch of functions for collision
Shape = Object:extend()
function Shape:new(x,y)
	self.x = x
	self.y = y
end
function Shape:getAngleTo(x,y)
	return math.atan2(y - self.y, x - self.x)
end
function Shape:getDistance(x,y)
	return math.sqrt(((self.x - x)^2) + ((self.y - y)^2))
end
function Shape:overlap(s2)
	return (self.x == s2.x and self.y == s2.y)
end
function Shape:setPos(x,y)
	self.x = x
	self.y = y
end
function Shape:draw()
	love.graphics.line(self.x,self.y,self.x,self.y)
end


--Derivatives of Shapes
-- Rectangle
Rectangle = Shape:extend()
function Rectangle:new(x,y,width,height)
	Rectangle.super.new(self,x,y)
	self.width = width
	self.height = height
end
function Rectangle:overlap(s2)
	if s2:is(Rectangle) then
		 return self.x + self.width > s2.x
			and self.x < s2.x + s2.width
			and self.y + self.height > s2.y
			and self.y < s2.y + s2.height
	elseif s2:is(Circle) then
		--Check the inner square in a circle 
		local check = {l = s2.x-(s2.radius * math.cos(45)),
			r = s2.x+(s2.radius * math.cos(45)),
			t = s2.y-(s2.radius * math.sin(45)),
			u = s2.y+(s2.radius * math.sin(45))}
		
		return self.x + self.width > check.l
			and self.x < check.r
			and self.y + self.height > check.t
			and self.y < check.u
	else
		return Rectangle.super.overlap(self,s2)
	end
end
function Rectangle:draw()
	love.graphics.rectangle("line",self.x,self.y,self.width,self.height)
end

-- Circle
Circle = Shape:extend()
function Circle:new(x,y,size)
	Circle.super.new(self,x,y)
	self.radius = size
end
function Circle:draw()
	love.graphics.circle("line",self.x,self.y,self.radius)
end
function Circle:overlap(s2)
	if s2:is(Rectangle) then 
		return s2:overlap(self)
	elseif s2:is(Circle) then
		local c = ((self.y-s2.y)^2) + ((self.x-s2.x)^2)
		return math.sqrt(c) < s2.radius + self.radius
	elseif s2:is(Shape) then
		local x = math.abs(self.x - s2.x) < self.radius
		local y = math.abs(self.y - s2.y) < self.radius
		return x and y
	end
	
end


--Collisions: Does physics collisions
--Takes in a rectangle and checks for overlaps. Based on the shape, will resolve collision
-- Maybe just extend Rectangle?
Collision = Rectangle:extend()
function Collision:new(x,y,width,height,strength)
	Collision.super.new(self,x,y,width,height)
	--Collsions
	if strength == nil then strength = 0 end
	self.strength = strength
	self.tempStr = strength
	self.last = {}
	self.last.x = x
	self.last.y = y
end
function Collision:setStrength(val)
	if val < 0 then val = 0 end
	self.strength = val
end
function Collision:update(dt)
	self.last.x = self.x
	self.last.y = self.y
	self.tempStr = self.strength
end
function Collision:check(col)
	if col:is(Collision) then return self:overlap(col) end
	return false
end
function Collision:collide(e,dir)
	if dir == "left" then
		local pushback = self.x + self.width - e.x
		self.x = self.x - pushback
	elseif dir == "right" then
		local pushback = e.x + e.width - self.x
		self.x = self.x + pushback
	elseif dir == "up" then
		local pushback = e.y + e.height - self.y
		self.y = self.y + pushback
	elseif dir == "down" then
		local pushback = self.y + self.height - e.y
		self.y = self.y - pushback
	end
end
function Collision:wasVerticalAligned(e)
	return self.last.y < e.last.y + e.height and self.last.y + self.height > e.last.y
end
function Collision:wasHorizontalAligned(e)
	 return self.last.x < e.last.x + e.width and self.last.x + self.width > e.last.x
end
--Overwritten by other functions
function Collision:checkResolve(e,dir)
	return true
end

function Collision:resolve(e)
	--Check if stronger; if true, reverse
	if self.tempStr > e.tempStr then
		return e:resolve(self)
	end
	--Check if colliding
	if self:check(e) then
		self.tempStr = e.tempStr
		if self:wasVerticalAligned(e) then
			if self.x + (self.width/2) < e.x + (e.width/2) then
				self:collide(e,"left")
			else
				self:collide(e,"right")
			end
			debugStrings[2] = "Horizontal Collision"
		elseif self:wasHorizontalAligned(e) then
			if self.y + (self.height/2) < e.y + (e.height/2) then
				self:collide(e,"down")
			else
				self:collide(e,"up")
			end
			debugStrings[2] = "Vertical Collision"
		end
		return true
	end
	return false
end

--TileMap: One tilemap layer. Takes a width, height, and a z index
-- Find a way to add animations.
-- Can be used for parallax as well.
-- Each tilemap layer contains a tile (tileId, relativex, relative y, rz = 0)
-- Each tile has an id, a cell x, and a cell y
-- Can be given a list of lists with data.
-- Also gives an array based on
-- Use Tiled with the right tileset to make layers, then export.
TileMap = Object:extend()
function TileMap:new(tileSetID,width,height,z,posX,posY)
	self.tileSet = BasicData.getTileSet(tileSetID)
	self.size = {w = width,h = height}
	if self.size.w <= 0 then self.size.w = 1 end
	if self.size.h <=0 then self.size.h = 1 end
	self.z = z
	self.pos = {x=posX,y=posY}
	--Setup the list
	self.map = {}
	for y = 1,height do
		local row = {}
		for x=1,width do table.insert(row,1) end
		table.insert(self.map,row)
	end
end
function TileMap:getRealPos(x,y)
	return {x * self.tileSet.tileSize.w, y * self.tileSet.tileSize.h}
end
function TileMap:getCell(x,y)
	if y > #self.map then return 0 end
	if x > #self.map[y] then return 0 end
	return self.map[y][x]
end
function TileMap:setCell(x,y,id)
	if y > #self.map then return end
	if x > #self.map[y] then return end
	self.map[y][x] = id
end
function TileMap:addToBatch(batch)
	local h = self.tileSet.tileSize.h
	local w = self.tileSet.tileSize.w
	for y,row in ipairs(self.map) do
		for x,tile in ipairs(row) do
			local q = self.tileSet:getTile(tile)
			batch:add(q,((x-1)*w)+self.pos.x,((y-1)*h)+self.pos.y)
		end
	end
end

--SoundObj:Contains a name, an audio file/source, and a serial number.
--Can set to loop or not. 
SoundObj = Object:extend()

function SoundObj:new(name,sourcePath,is_sfx)
	self.name = name	
	self.source = nil
	self.type = ""
	local stype = ""
	if is_sfx then
		self.type = "sfx"
		stype = "static"
	else
		self.type = "music"
		stype = "stream"
	end
	if love.filesystem.getInfo(sourcePath) then
		self.source = love.audio.newSource(sourcePath,stype)
	end
	
end
function SoundObj:play()
	if self.source ~= nil then self.source:play() end
end
function SoundObj:stop()
	if self.source ~= nil then self.source:stop() end
end
function SoundObj:pause()
	if self.source ~= nil then self.source:pause() end
end
function SoundObj:setLoop(b)
	if self.source ~= nil then self.source:setLooping(b) end
end

--SoundPlayer: Contains a list of sound indexes. Plays and stops them on command
-- Can add or remove sound effects.
-- Is being refectored to work ith 


--ENTITIES
--Entity: an abstract class for the majority of interactable game objects
-- Mostly acts as a vehicle for the Basic Process and Basic draw
-- 
Entity = Object:extend()
BasicData.entities = {}
function Entity:new(x,y,z)
	self.x = x
	self.y = y
	self.z = z
	--Drawing
	self.img = nil
	self.drawCol = false
	self.drawImg = true
	--Processing
	self.process_order = 0
	table.insert(BasicData.entities,self)
	self.id = #BasicData.entities
	self.kill_this = false
	self.paused = false
	self.can_process = true
	--Movement; this movement is tracked to delta
	self.dx = 0
	self.dy = 0
end
function Entity:setDelta(x,y)
	self.dx = x
	self.dy = y
end
function Entity:setImg(img)
	self.img = img
end
function Entity:kill()
	self.kill_this = true
end
function Entity:pause(b)
	if b == nil then b = self.pause end
	self.paused = b
end
function Entity:setDraw(canDraw,drawCol)
	if canDraw == nil then canDraw = true end
	if drawCol == nil then drawCol = false end
	self.drawImg = canDraw
	self.drawCol = drawCol

end
function Entity:process(dt)
	--Img
	if self.img == nil then
		--Code here
	elseif self.img:is(Animation) then
		self.img:update(dt)
	end
	--Normalize dx dy

	--Movement
	self.x = self.x + (self.dx *dt)
	self.y = self.y + (self.dy *dt)
end
function Entity:drawCollision()
	love.graphics.rectangle("line",self.x,self.y,4,4)
end
function Entity:draw()
	--Draw the image
	if self.img == nil then
		--Code here
	elseif self.img:is(Animation) then
		if self.drawImg then self.img:draw(self.x,self.y) end
	end
	--Draw the collision
	if self.drawCol then self:drawCollision() end
end
function Entity:overlap(e)
	if e:is(Entity) then return (self.x == e.x and self.y == e.y) end
	return false
end
function Entity:collision(e)
	return false
end
function BasicData.process(dt)
	--Sort all entities
	--Process order first; otherwise, id
	table.sort(BasicData.entities,function(e1,e2)
		if e1.process_order ~= e2.process_order then 
			return e1.process_order > e2.process_order 
		end
		return e1.id < e2.id
	end)
	--Go through
	for i,e in ipairs(BasicData.entities) do
		if e:is(Entity) then 
			if e.kill_this == true then
				table.remove(BasicData.entities,i)
				i = i - 1
			elseif e.paused ~= true and e.can_process then 
				e:process(dt) 
			end
		end
	end
end
function BasicData.draw()
	--Sort all entities
	-- Higher Z > Higher Y > Lower Y > Lower Z
	table.sort(BasicData.entities,function(e1,e2)
		if e1.z ~= e2.z then 
			return e1.z < e2.z			
		end
		if e1.y ~= e2.y then 
			return e1.y < e2.y			
		end
		return e1.id < e2.id
	end)
	--Go through
	for i,e in ipairs(BasicData.entities) do
		--Later check if on camera before drawing
		--Draw for now
		if e:is(Entity) then 
			e:draw() 
		end
	end
end


-- Actors : Have a collision, and mostly move around.
-- All Actors are added to BasicData objects
-- They will collide with walls (not added yet)
BasicData.actors = {}
Actor = Entity:extend()

function Actor:new(x,y,z,priority)
	Actor.super.new(self,x,y,z)
	self.col = nil
	self.priority = priority
	table.insert(BasicData.actors,self)
	self.id = #BasicData.actors
	self.canCollide = false
end
function Actor:setCollision(width,height,str)
	self.col = Collision(self.x,self.y,width,height,str)
end
function Actor:canCollide(b)
	if b == nil then b = false end
	self.canCollide = b
end
function Actor:setPriority(i)
	self.priority = Lume.clamp(i,0,100)
end

function Actor:process(dt)
	Actor.super.process(self,dt)
	if self.col ~= nil then
		if self.col:is(Collision) then
			self.col.x = self.x
			self.col.y = self.y
			self.col:update(dt)
		end
	end
end
function Actor:drawCollision()
	if self.col == nil then
		Actor.super.draw(self)
	elseif self.col:is(Collision) then
		self.col:draw()
	end
end

function Actor:hasCollision()
	if self.col == nil then return false end
	if self.col:is(Collision) then return true end
	return false
end
function Actor:collide(a)
	if a == nil then return false end
	if a:is(Actor) then
		if a:hasCollision() then 
		--Resolve
		local t = self.col:resolve(a.col) 
		--Set position
			if t then
				self.x = self.col.x
				self.y = self.col.y
				a.x = a.col.x
				a.y = a.col.y
			end
		return t
		end
	end
	return false
end

--This should be way later.
function BasicData.collisions(dt,loops)
	if loops == nil then loops = 100 end
	
	--Sort in priority order
	table.sort(BasicData.actors,function(a1,a2)
		if a1.priority ~= a2.priority  then
			return a1.priority > a2.priority
		end
		return a1.id < a2.id
	end)
	
	--Check collide
	local can_loop = true
	local limit = 0
	while can_loop do
		--Breaking the loop
		can_loop = false
		limit = limit + 1
		if limit >= loops then break end
		--Collisions
		for i=1,#BasicData.actors-1 do
			for j=i+1,#BasicData.actors do
				local col = BasicData.actors[i]:collide(BasicData.actors[j])
				if col then can_loop = true end
			end
		end
	end
end



return BasicData